#######################################################################
# File Name    : proj_main.tcl
# Date Created : 12/01/99 
# Author       : Tim L. Wilson, Gregg D. Lahti & Rodney Pesavento
# Project      : TOPS Example Code for SNUG
#######################################################################
# $Id$
#######################################################################
# Description:
# This script defines the project synthesis flow. It is intended to 
# provide a generic synthesis capability for most synthesis targets.
# For targets that require more specific synthesis flows, the 
# target_main.tcl may be sourced from this script.
#######################################################################

#######################################################################
# Check for required variables
#######################################################################
set dc_shell_status [catch {list $G_DESIGN_FILE_NAME}];
if {$dc_shell_status != 0} {
  echo "#MAIN: Error-Variable G_DESIGN_FILE_NAME not set";
  quit; # Fatal Error, exit the tool 
}
set dc_shell_status [catch {list $G_SUBFUBS}];
if {$dc_shell_status != 0} {
  echo "#MAIN: Info G_SUBFUBS not set, setting to null";
  set G_SUBFUBS "";
}

# Set G_DESIGN_NAME (remove path & extension)
set G_DESIGN_NAME [file tail [file rootname $G_DESIGN_FILE_NAME]];

# Save start time
set G_MAIN_START_TIME [clock seconds];
P_timestamp;

#######################################################################
# Source the Project, Local & Target specific procedures
#######################################################################
P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_dc_setup.tcl] "MAIN";
P_source_if_exists [file join ${G_PROJ_SYN} proj_procs.tcl] "MAIN";
P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_procs.tcl] "MAIN";
P_source_if_exists [file join ${G_PROJ_SYN} report_procs.tcl] "MAIN";
P_source_if_exists [file join ${G_PROJ_SYN} proj_reports.tcl] "MAIN";

#######################################################################
# Load project library setup script
#######################################################################
P_source_if_exists [file join ${G_PROJ_SYN} proj_library.tcl] "MAIN";

#######################################################################
# Source the Target Main script which overrides TOPS Main.  This allows
# project to define special flows. The flow control should exit in that
# path.
#######################################################################
P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_main.tcl] "MAIN";

echo ""
echo "############################################################"
echo "#        Design Analyze & Elaborate Phase"
echo "############################################################"
echo ""

# Set timestamp & Read the SUBFUB's if they exist
set G_TIME_MARK [clock seconds];
P_timestamp;

#######################################################################
# Analyze ONLY the SUBFUB list
#######################################################################
P_read_subfubs $G_SUBFUBS;

#######################################################################
# Analyze & elaborate top level design 
#######################################################################
P_read_top $G_DESIGN_FILE_NAME;

#######################################################################
# Get the Id CVS string from the source files
#######################################################################
set G_SRC_ID_LIST [P_get_srcline $G_DESIGN_FILE_NAME $G_SUBFUBS];

# Link design
link;

echo "#MAIN: Writing out unmapped db for $G_DESIGN_NAME";
write -f db -hier -o [file join ${G_DBS_PATH} ${G_DESIGN_NAME}_unmapped.db] $G_DESIGN_NAME;

echo "#MAIN: Elapsed time of Analyze & Elaborate Phase was" [P_elapsed_time $G_TIME_MARK];

echo ""
echo "############################################################"
echo "#        Apply Synthesis Design Constraints Phase"
echo "############################################################"
echo ""

set G_TIME_MARK [clock seconds];
P_timestamp;

#######################################################################
# Set the operating conditions
#######################################################################
P_proj_operating_conditions;

#######################################################################
# Set wire_load. Design-specific script overrides project default
#######################################################################
# Source Target-level, else set the default project WLM
if {[P_exec_if_exists P_${G_DESIGN_NAME}_wire_load "MAIN"]} {
} elseif {[P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_wire_load.tcl] "MAIN"]} {
} else {P_proj_wire_load}

#######################################################################
# Project & Target-level constraint loop.
# Include clocks/resets & constraints.
# Turn off auto-link to speed up constraint reading
#######################################################################
set auto_link_disable true;

# Set clocks for design
if {[P_exec_if_exists P_${G_DESIGN_NAME}_clocks "MAIN"]} {
} elseif {[P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_clocks.tcl] "MAIN"]} {
} else {P_proj_clocks}

# Set Constraints for design. This list could be extended if needed
set includes {constraints};
foreach file $includes {
  # Target-level files REPLACE project-level files
  if {[P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_${file}.tcl] "MAIN"]} {
  } else {P_source_if_exists [file join ${G_PROJ_SYN} proj_${file}.tcl] "MAIN"}
}

#######################################################################
# Target-level only constraint loop for dont_use & exceptions.
# Design-specific scripts are added to project defaults
#######################################################################
# Create a variable with for the library without the corner
set lib [lindex [split $G_LIBRARY_NAME "_"] 0];
# This list could be extended if needed
set includes "{${lib}_dont_use} {exceptions}";
foreach file $includes {
  # Source target constraints if they exist
  P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_${file}.tcl] "MAIN"
}

set auto_link_disable false;

echo "#MAIN: Elapsed time of Constraints Phase was" [P_elapsed_time $G_TIME_MARK];

#######################################################################
# Check Timing Constraints 
#######################################################################
check_timing;

echo ""
echo "############################################################"
echo "#        Synthesis Compilation Phase"
echo "############################################################"
echo ""

set G_TIME_MARK [clock seconds];
P_timestamp;

# If Target Compile exists, then source it, otherwise do Project Compile
if {[P_exec_if_exists P_${G_DESIGN_NAME}_compile "MAIN"]} {
} elseif {[P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_compile.tcl] "MAIN"]} {
} else {P_proj_compile}

echo "#MAIN: Elapsed time of Compile Phase was" [P_elapsed_time $G_TIME_MARK];

set G_TIME_MARK [clock seconds];
P_timestamp;

#######################################################################
# Change the names before writing out the ouptuts
#######################################################################
P_proj_change_names $G_LOG_PATH $G_DESIGN_NAME;

echo "#MAIN: Elapsed time of Change Names was" [P_elapsed_time $G_TIME_MARK];

echo ""
echo "############################################################"
echo "#        Write Design Outputs Phase (db, netlist, SDF, SDC)"
echo "############################################################"
echo ""

set G_TIME_MARK [clock seconds];
P_timestamp;

# If Target Outputs exists, then exec it, otherwise do Project Outputs
if {[P_exec_if_exists P_${G_DESIGN_NAME}_outputs "MAIN"]} {
} elseif {[P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_outputs.tcl] "MAIN"]} {
} else {P_proj_outputs $G_DBS_PATH $G_DESIGN_NAME $G_OUTPUTS_PATH $G_SRC_ID_LIST}

echo "#MAIN: Elapsed time of Output Generation Phase was" [P_elapsed_time $G_TIME_MARK];

echo ""
echo "############################################################"
echo "#        Write Design Reports Phase"
echo "############################################################"
echo ""

set G_TIME_MARK [clock seconds];
P_timestamp;

# If Target Reports exists, then exec it, otherwise do Project Reports
if {[P_exec_if_exists P_${G_DESIGN_NAME}_reports "MAIN"]} {
} elseif {[P_source_if_exists [file join ${G_SCRIPTS_PATH} ${G_DESIGN_NAME}_reports.tcl] "MAIN"]} {
} else {P_proj_reports $G_REPORTS_PATH $G_DESIGN_NAME}

echo "#MAIN: Elapsed time of Report Generation Phase was" [P_elapsed_time $G_TIME_MARK];

echo "#MAIN: Elapsed time of proj_main script was " [P_elapsed_time $G_MAIN_START_TIME];
P_timestamp;

#######################################################################
# Filter the synthesis compile log for Errors, Warnings, & Elapsed time
#######################################################################
#
# GDL 08-12-00
# Removed, does not work reliably due to I/O buffering from DC and
# variances in UNIX OS's.  Moved to synbatch instead but kept the 
# placeholder here for future reference.
# 
# P_log_filter $G_LOG_PATH $G_DESIGN_NAME;

exit 0;

#######################################################################
# $Log$
#######################################################################



