#######################################################################
# File Name    : proj_procs.tcl
# Date Created : 3/06/2000 
# Author       : Tim L Wilson, Gregg D. Lahti & Rodney Pesavento
# Project      : TOPS Example Code for SNUG
#######################################################################
# $Id$
#######################################################################
# Description: Project-Specific utility procedures for synthesis
#######################################################################

#######################################################################
# Usage: P_proj_annotate_clocks {min_uncertainty}
#
# This procedure removes insertion delay & prepares clocks for postlayout analysis
#######################################################################
proc P_proj_annotate_clocks {min_uncertainty} {   
  # Adjust the clock uncertainty to min
  echo "Annotating all defined clocks for postlayout analysis";
  set_clock_uncertainty 10 [all_clocks];
  set_clock_latency 0.0 -source [all_clocks];

  # Do not want this clock to be ideal, set to propagated
  set_propagated_clock [all_clocks];
}; # end P_annotate_clocks

# This is the help attribute definition
define_proc_attributes P_proj_annotate_clocks \
  -info "PROJ_PROCS: Removes insertion delay & prepares clocks for post-layout analysis" \
  -define_args {
    {min_uncertainty "Minimum Clock uncertainty" min_uncertainty float required}
  }

#######################################################################
# Usage: P_proj_clocks 
#
# This procedure is used to setup the projects clocks.
#######################################################################
proc P_proj_clocks {} {
  global DEFAULT_CLOCK;

  # Define all clocks with the following invocation format:
  #
  # P_proj_constrain_clock clk_name period uncertainty insertion
  #
  #   REQUIRED arguments:
  #   clk_name:    clock name
  #   period:      clock period (time units of lib ps)
  #   uncertainty: clock uncertainty (time units of lib ps)
  #   insertion:   clock insertion delay (time units of lib ps) 
  #
  #   OPTIONAL arguments (the following should only be necessary when manipulating
  #                       duty cycle and complex waveform creation):
  #   hierarchy:   hierarchical instance path to clock
  #   rise1:       first rise time (time units of lib ps) 
  #   fall1:       first fall time (time units of lib ps) 
  #   rise2:       next rise (time units of lib ps) 
  #   fall2:       nest fall (time units of lib ps) 
  #
  #  Example:
  #      P_proj_constrain_clock "clk_unit" 5.024 .175 0.250 2.762;

  # Project clocks
  P_proj_constrain_clock "clk" 3000 100 50;

  # Setup default in case there is only one clock
  set G_DEFAULT_CLOCK pclk;

  ###################################################################
  # Create a clock for multicycle path timing exceptions, ir required
  ###################################################################
  #create_clock -period 38000 -name apb_pclk_mc;

  # Display current clock info 
  report_clock -skew -attributes -nosplit;

  ###################################################################
  # Constrain reset ports for design
  ###################################################################
  P_proj_constrain_reset "reset_n";

}; # end P_proj_clocks

define_proc_attributes P_proj_clocks \
  -info "PROJ_PROCS: Procedure to setup all the project clocks"

#######################################################################
# Usage: P_proj_constrain_reset {reset_name}
#
# This procedure is used to set the constraints on the "reset" ports
#######################################################################
proc P_proj_constrain_reset {reset_name} {   
  global G_TOPS;
  global synopsys_program_name;

  # If port does not exist, send warning results to bit bucket
  redirect $G_TOPS(NULL) \
     {set test4rst [get_ports -filter {@port_direction == in} $reset_name]};

  # If port exists, go ahead and constrain
  if {$test4rst != {}} {
    echo "#RESET-Info: Setting reset attributes on port - $reset_name";

    remove_driving_cell $reset_name;
    set_ideal_net $reset_name;
    set_drive 0 $reset_name;
    set_false_path -from $reset_name;

    # This attribute gets used in the constraints script
    if {[set synopsys_program_name] != "pt_shell"} {
      set_attribute $reset_name is_reset true -type boolean;
    }
  }; # end if
  unset test4rst;
}; # end P_proj_constrain_reset

define_proc_attributes P_proj_constrain_reset \
  -info "PROJ_PROCS: Procedure to set the constraint on RESET port" \
  -define_args {
    {reset_name "Name of the port on which constraints are to be set" reset_name string required}
}

#######################################################################
# Usage: P_proj_constrain_clock {clk_name period uncertainty insertion \
#                                  hier {r1 ""} {f1 ""} {r2 ""} {f2 ""}}
# Example: P_proj_constrain_clock {pciclk 1000.0 150.0 0 "" 0 400 0 0}
#
# This procedure is used to define & create clock "clk_name" with attributes.
# NOTE: User must specify "hier"/ if setting waveform vars
#######################################################################
proc P_proj_constrain_clock {clk_name period uncertainty insertion {hier ""} {r1 ""} {f1 ""} {r2 ""} {f2 ""}} {
  global G_CLOCK_ARRAY;
  global G_TOPS;
  global G_CWLM_LIBRARY;
  global synopsys_program_name;

  # Test for the clock to exist. use hierarchy if supplied  
  if {$hier != ""} {
     set clk_name $hier$clk_name;
     # test4clk is a collection, if the collection exits the clock exists
     redirect $G_TOPS(NULL) \
        {set test4clk [get_ports $clk_name]};
  } else {
     # test4clk is a collection, if the collection exits the clock exists
     redirect $G_TOPS(NULL) \
        {set test4clk [get_ports -filter {@port_direction == in} $clk_name]};
  }

  # Constrain clocks if they exist
  if {$test4clk != {}} {
    echo "#CLOCKS-Info: Setting clock attributes on port - $clk_name";

    # Create a global array to contain all clock info
    set G_CLOCK_ARRAY($clk_name,name) $clk_name;
    set G_CLOCK_ARRAY($clk_name,period) $period;
    set G_CLOCK_ARRAY($clk_name,uncertainty) $uncertainty;
    set G_CLOCK_ARRAY($clk_name,insertion) $insertion;

    # Create clock
    if {$r1 != "" || $f1 != "" || $r2 != "" || $f2 != ""} {
       create_clock -period $period -waveform [list $r1 $f1 $r2 $f2] $clk_name;
    } else {
       create_clock -period $period $clk_name;
    }

    # Only if not hierarchical
    if {$hier == ""} {
       remove_driving_cell $clk_name;

       # Set input port drive to infinite drive capability (no delay)
       set_drive 0 $clk_name;
    }

    # Sets the dont_touch attribute on pin or clock
    set_dont_touch_network $clk_name;

    # Set the uncertainty, models clock skew
    set_clock_uncertainty $uncertainty $clk_name;

    # Sets insertion delay of clock tree
    set_clock_latency $insertion -source $clk_name;

    # This attribute gets used in the constraints script
    if {[set synopsys_program_name] != "pt_shell"} {
      set_attribute $clk_name is_clock true -type boolean;
    }

    # Only use cwlm if it exists
    if {$G_CWLM_LIBRARY != "{}" && $hier == ""} {
      # Get cwlm rootname & set_wire_load
      set cwlm_lib_name [file rootname $G_CWLM_LIBRARY];
      set_wire_load_model -name CLKNET -lib $cwlm_lib_name $clk_name;
    } else {
      set_resistance 0.0 $clk_name;
      set_load 0.0 $clk_name;
    }; # end if
  }; # end if
  unset test4clk;
  return;
}; # end P_proj_constrain_clock

# This is the help attribute definition
define_proc_attributes P_proj_constrain_clock \
  -info "PROJ_PROCS: Define & create clock for clk_name with attributes" \
  -define_args {
    {clock_name "Clock Name" clock_name string required}
    {period "Clock Period in library time units" period float required}
    {uncertainty "Clock Uncertainty in library time units" uncertainty float required}
    {insertion "Clock Insertion Delay in library time units" insertion float required}
    {hier "Hierarchy to the clock source" hier string}
    {r1 "Waveform Rise1" r1 float}
    {f1 "Waveform Fall1" f1 float}
    {r2 "Waveform Rise2" r2 float}
    {f2 "Waveform Fall2" f2 float}
  }

#######################################################################
# Usage: P_proj_inst_cnt
#
# This procedure is used to count the number of instances
#######################################################################
proc P_proj_inst_cnt {} {
  set cell_count [sizeof_collection [get_cells -hier  * ]]

  set leaf_cells [get_cells -hier * -filter "@is_hierarchical == false"]

  set hier_cell_count [sizeof_collection [get_cells -hier * -filter "@is_hierarchical == true"]]
  set leaf_cell_count [sizeof_collection [get_cells -hier * -filter "@is_hierarchical == false"]]

  set seq_leaf_count [sizeof_collection [filter $leaf_cells "@is_sequential == true"] ]
  set combo_leaf_count [sizeof_collection [filter $leaf_cells "@is_combinational == true"]]

  echo "************************************************"
  echo "* "
  echo "* Design              : "  [get_object_name [current_design]]
  echo "* Number of cells     : "  $cell_count  
  echo "*    Hierarchical     : "  $hier_cell_count 
  echo "*    Leaf             : "  $leaf_cell_count  
  echo "*       Sequential    : "  $seq_leaf_count
  echo "*       Combinational : "  $combo_leaf_count
  echo "* "
  echo "************************************************"
  return 0;
}; # end P_proj_cell_cnt

define_proc_attributes P_proj_inst_cnt \
  -info "PROJ_PROCS : Procedure to count the number of instances"

##########################################################################
# Usage: P_proj_change_names
#
# This procedure defines the projects change names rules (removes hierarchy separators 
# and special characters from netlists) and is included prior to the outputs
# being written out.
# 
# Magic variables - for retaining busses
# Note: these are only a few of the variables used to write out
# HDL and EDIF. The others have their default values in this example.
##########################################################################
proc P_proj_change_names {log_path_lcl design_name_lcl} {
  # Set global for scoping
  global bus_inference_style;
  global bus_naming_style;
  global change_names_dont_change_bus_members;
  global hdlout_internal_busses

  echo "#PROJ: Changing project names";
  
  ##########################################################################
  # Name rule to convert "/" (hierarchical separator) to "_"
  # Note: Since bus_naming_style is changed you will see NMA-16 warnings.
  # These are OK; the busses will still be preserved.
  # Also, a separate rule for the slash makes the output more readable.
  # the simple_??? rule fixes / but not as cleanly as a separate rule.
  # slashes occur in designs when they are ungrouped.
  ##########################################################################
  
  define_name_rules slash -restricted {/} -replace {_};
  
  ##########################################################################
  # Name rule to make legal VHDL, Verilog names. Note: This example
  # only covers one-dimensional arrays. For two-dimensional arrays the
  # mapping should have four pairs instead of two.
  ##########################################################################
  # There are different name rules for ports and nets vs cells.
  # removing the bus naming style causes mismatches between ports and nets.
  # NOTE: it is dangerous to change the port names, mismatches between
  # connected instances can occur.
  ##########################################################################
  # Top level Ports and Nets connected to Ports must already follow the naming 
  # convention, however Synopsys created nets and ports (think DW) may not.
  ##########################################################################
  
  define_name_rules simple_port \
    -allowed {A-Za-z0-9[]_} -first {0-9_} -last {_} \
    -equal_ports_nets -type port;
  
  define_name_rules simple_net \
    -allowed {A-Za-z0-9[]_*} -first {0-9_} -last {_} \
    -type net;

  define_name_rules simple_cell \
    -allowed {A-Za-z0-9_} -first {0-9_} -last {_} \
    -map {{ {"[","_"} {"]",""} {"[","_"} {"]",""} }} -remove -type cell;
  
  ##########################################################################
  # Exec desired names rules
  ##########################################################################
  redirect [file join ${log_path_lcl} ${design_name_lcl}_change_names_slash.log] \
    {change_names -hier -rules slash -verb};
  
  redirect [file join ${log_path_lcl} ${design_name_lcl}_change_names_simple.log] \
    {report_name_rules simple_port};
  redirect -append [file join ${log_path_lcl} ${design_name_lcl}_change_names_simple.log] \
    {change_names -hier -rules simple_port -verb};

  redirect -append [file join ${log_path_lcl} ${design_name_lcl}_change_names_simple.log] \
    {report_name_rules simple_net};
  redirect -append [file join ${log_path_lcl} ${design_name_lcl}_change_names_simple.log] \
    {change_names -hier -rules simple_net -verb};
  
  redirect -append [file join ${log_path_lcl} ${design_name_lcl}_change_names_simple.log] \
    {report_name_rules simple_cell};
  redirect -append [file join ${log_path_lcl} ${design_name_lcl}_change_names_simple.log] \
    {change_names -hier -rules simple_cell -verb};
}; # end P_proj_change_names

define_proc_attributes P_proj_change_names \
  -info "PROJ_PROCS: Procedure to remove hierarchy separators and special characters \ 
netlists and SDF's, included prior to netlist being written out." \
  -define_args {
     {log_path_lcl "Path to the log files" log_path_lcl string required}
     {design_name_lcl "design name" design_name_lcl string required}
}

#######################################################################
# Usage: P_proj_compile 
#
# This procedure is used to create the project compilation strategy
# May include procedures for pre & post scan insertion
#######################################################################
proc P_proj_compile {} {
  set_flatten false;
  set_structure true;

  #######################################################################
  # May want to try new boolean optimization for 1997.01 & up
  #######################################################################
  # set compile_new_boolean_structure true
  # set_structure true -boolean true -boolean_effort medium

  # Fix multiply instantiated designs
  uniquify;

  # This is required for test-ready compile (compile -scan)
  P_get_license "Test-Compiler";

  echo "#PROJ: Performing project compile";
  compile -map medium -scan;

  # Ungroup the designware components
  P_proj_ungroup_dw;

  # Perform incremental compile to clean up after the DW ungrouping
  compile -map_effort low -incremental;
}; # end P_proj_compile

define_proc_attributes P_proj_compile \
  -info "PROJ_PROCS: compiles the project"

#######################################################################
# Usage: P_proj_compile_top 
#
# This procedure is used to compile the top-level netlist or the top
# of any target (unit, super-unit) that contains modules that do not
# need to be compiled.
#######################################################################
proc P_proj_compile_top {} {
  set_flatten false;
  set_structure true;

  echo "#PROJ: Performing project top-level  compile";
  compile -top;
}; # end P_proj_compile_top

define_proc_attributes P_proj_compile_top \
  -info "PROJ_PROCS: compiles the project top-level"

#######################################################################
# Usage: P_proj_insert_buffers lib_cell_name inst_string fix_hold_mode
#
# This procedure is used to fix HOLD violations by inserting strategic
# buffers in the scan, data or both paths of the design. It is intended
# to be invoked pre-layout to reduce the number of violations.
#######################################################################
proc P_proj_insert_buffers {lib_cell_name inst_string fix_hold_mode} {

  # Generate list of violators & filter on VIOLATED
  report_constraint -min_delay -all > tmp;
  set violator_list [exec sed -n -e "/VIOLATED/p" tmp];
  sh rm tmp;

  # Check for no violations
  if {$violator_list == ""} {
    echo "#INFO: No MIN violations in current design";
    return;    
  }; # end if

  set si_violator_list "";
  set d_violator_list "";

  # Now filter on "/SI" & create list
  foreach violator $violator_list {
    if [string match */SI $violator] {
#      echo $violator;
      set si_violator_list [concat $si_violator_list $violator];
    }; # end if
    if [string match */D $violator] {
#      echo $violator;
      set d_violator_list [concat $d_violator_list $violator];
    }; # end if
  }; # end foreach

  # Build the list to fix
  if {$fix_hold_mode == "scan"} {
    set all_violator_list $si_violator_list;
  } elseif {$fix_hold_mode == "data"} {
    set all_violator_list $d_violator_list;
  } elseif {$fix_hold_mode == "both"} {
    set all_violator_list [concat $d_violator_list $si_violator_list];
  } else {
      echo "ERROR: Incorrect Fix Hold Mode. Please specify scan, data or both";
  }; # end if

  # Reset cell counter (used to create a unique instance name
  set cell_cnt 0;
  
  echo "#INFO: Fixing Hold violations on the violator paths.";
  
  # Fix all the violations for the violator list
  foreach reg_cell $all_violator_list {
  
    # Get the instance name of the violator
    regexp {(.*)/.*} $reg_cell var1 var2
    set inst_name [file tail $var2];
  
    # Get the hierarchy path to the violating register
    regexp {(.*)/.*} $var2 var1 hier_path
  
    # Increment the cell counter
    set cell_cnt [expr $cell_cnt + 1];

    # Find everything connected to the pin of this register
    set orig_net [all_connected [get_pins $reg_cell]];
    # Create new instance name (HOLD delay cell)
    set new_cell ${inst_name}${inst_string}${cell_cnt};
    # Create new net name (net inserted between new HOLD cell & register pin)
    set new_net ${new_cell}_net;
    
    # Disconnect the orig net & create the new cell & net
    disconnect_net $orig_net $reg_cell;
    create_cell -instance $hier_path $new_cell $lib_cell_name;
    create_net -instance $hier_path $new_net;
  
    # Add the hierarchy path to the new net & cell vars
    set new_net [file join $hier_path $new_net];
    set new_cell_in [file join $hier_path ${new_cell}/A];
    set new_cell_out [file join $hier_path ${new_cell}/O];
  
    # Connect the nets
    connect_net $new_net $reg_cell;
    connect_net $new_net $new_cell_out;
    connect_net $orig_net $new_cell_in;
  }; # end foreach

  echo "#INFO: Added $cell_cnt HOLD cells to the design.";
}; # end proc 

define_proc_attributes P_proj_insert_buffers \
  -info "PROJ: Procedure to fix HOLD violations inserting strategic buffers" \
  -define_args {
    {lib_cell_name "Library Cell Name to insert" lib_cell_name string required}
    {inst_string "Instance String to insert in the new cell" inst_string string required}
    {fix_hold_mode "Fix Hold Mode: scan, data, or both" fix_hold_mode string required}
}


#######################################################################
# Usage: P_proj_operating_conditions 
#
# This procedure Sets default operating conditions using min/max conditions
# from library.
# NOTE: Synopsys does not maintain this information when saving a design.
#       This proc needs to be run again when loading a design into memory.
#######################################################################
proc P_proj_operating_conditions {} {
  global G_MAX_OP_CONDITION;
  global G_MIN_OP_CONDITION;
  global G_MAX_LIBRARY;
  global G_MIN_LIBRARY;

  # Set min_library only if G_MAX_LIBRARY not same as G_MIN_LIBRARY
  if {![string match ${G_MAX_LIBRARY} ${G_MIN_LIBRARY}]} {
    # Assume that min library already in memory (proj_library.tcl)
    set_min_library ${G_MAX_LIBRARY}.db -min_version ${G_MIN_LIBRARY}.db;
  }; # end if

  # Set operating conditions
  set_operating_conditions \
     -max $G_MAX_OP_CONDITION \
     -max_library $G_MAX_LIBRARY \
     -min $G_MIN_OP_CONDITION \
     -min_library $G_MIN_LIBRARY;

}; # end of P_proj_operating_conditions

define_proc_attributes P_proj_operating_conditions \
  -info "PROJ_PROCS: Sets the default operating conditions using the min/max conditions from the library"

#######################################################################
# Usage: P_proj_precompile_scan {num_of_chains}
#
# This procedure is used to set the scan config & define the scan ports
# for the design.
#######################################################################
proc P_proj_precompile_scan {num_of_chains} {
  echo "#PROJ: Setting scan configuration";

  # Provided the full spectrum of scan options, choose wisely
  set_scan_configuration \
    -add_lockup false \
    -area_critical false \
    -bidi_mode input \
    -chain_count $num_of_chains \
    -clock_gating entire_design \
    -clock_mixing no_mix \
    -dedicated_scan_ports true \
    -disable false \
    -existing_scan false \
    -hierarchical_isolation false \
    -methodology full_scan \
    -multibit_segments true \
    -rebalance false \
    -replace true \
    -route true \
    -route_signals "serial scan_enables" \
    -style multiplexed_flip_flop \
    -internal_clocks true \
    -create_test_clocks_by_system_clock_domain false \
    -physical false

  # Define naming style for the design after inserting scan
  set insert_test_design_naming_style "%s_test_%d";

  # Set the scan test mode pins
  #set_test_hold 1 [get_port scantestmode];

  # Method 1: Allow Synopsys to determine scan ports
  # Here are some variables to rename the ports used for scan chains
  set test_scan_enable_port_naming_style "test_se%s";
  set test_scan_in_port_naming_style "test_si%s%s";
  set test_scan_out_port_naming_style "test_so%s%s";

  # Method 2: Specify the scan ports for Synopsys
  #set_scan_path chain1
  #set_scan_signal test_scan_in -port my_scan_in -chain chain1
  #set_scan_signal test_scan_out -port my_scan_out -chain chain1

}; # end P_proj_precompile_scan

define_proc_attributes P_proj_precompile_scan \
  -info "PROJ_PROCS: Procedure to set scan configuration"

######################################################################
# Usage: P_proj_postcompile_scan {report_path design_name}
#
# This procedure is used to insert scan
#######################################################################
proc P_proj_postcompile_scan {report_path design_name} {
  echo "#PROJ: Performing check_test";
  P_report_check_test $report_path $design_name;

  echo "#PROJ: Inserting scan chains";

  # Keeps insert_scan from placing inverters in scan path
  set test_disable_find_best_scan_out true;
  set test_dont_fix_constraint_violations true;

  # Insert scan chains based on variables set
  #insert_scan;
  insert_scan -ignore_compile_design_rules;

}; # end P_proj_postcompile_scan

define_proc_attributes  P_proj_postcompile_scan \
  -info "PROJ_PROCS : Procedure to insert scan"

##########################################################################
# Usage: P_proj_outputs {dbs_path design_name outputs_path srcline \
#                       {out_string ""} {out_type ""}}
#
# This procedure generates the default outputs. This includes db, sdf, sdc &
# verilog gate netlist.
# srcline is a list.
# NOTE: If out_string is null, the proc will only output db & verilog netlist.
#       If out_type is "layout", all layout related info is written.
##########################################################################
proc P_proj_outputs {dbs_path design_name outputs_path srcline {out_string ""} {out_type ""}} {
  global bus_naming_style;
  global hdlout_internal_busses;
  global set_fix_multiple_port_nets;
  global verilogout_single_bit;

  #######################################################################
  # Write a mapped DB file
  #######################################################################
  # By default, out_string = "". If USER requires other name, set $out_string
  set output_file [file join ${dbs_path} ${design_name}${out_string}.db];

  echo "#OUTPUTS: Writing out mapped db to ${output_file}";
  write -f db -hier -o $output_file;

  #######################################################################
  # Set naming style comments (VHDL or Verilog)
  #######################################################################
  set timestamp [exec date];
  
  #######################################################################
  # Comments for verilog and sdf
  #######################################################################
  set vlog_comment "//";
  set vlog_info "$vlog_comment This netlist was created with the following source files:";
  set vlog_timestamp "$vlog_comment $timestamp";
  set vlog_cvsline "$vlog_comment \$Id\$";
  
  #######################################################################
  # Verilog Gates
  #######################################################################
  if {$out_string == ""} {
     set out_string "_gates"
  }

  set output_file [file join ${outputs_path} ${design_name}${out_string}.v];
  echo "#OUTPUTS: Writing verilog netlist to ${output_file}";
  P_get_license "HDL-Compiler";
  
  write -f verilog -hier -o $output_file;
  
  redirect -append $output_file {echo $vlog_info};
  redirect -append $output_file {echo $vlog_timestamp};

  # Dump the CVS ID's for all elements in the srcline list
  foreach srcline_elem $srcline {
    echo $srcline_elem;
    set vlog_srcline "$vlog_comment netlist CVS srcid: $srcline_elem";
    redirect -append $output_file {echo $vlog_srcline};
  }; # end foreach
  redirect -append $output_file {echo $vlog_cvsline};

  remove_license "HDL-Compiler";
    
  #######################################################################
  # Write SDF file
  #######################################################################
  set output_file [file join ${outputs_path} ${design_name}${out_string}.max.sdf];
  echo "#OUTPUTS: Writing sdf timing data to ${output_file}";
  write_sdf -version 2.1 $output_file;
  
  redirect -append $output_file {echo $vlog_info};
  redirect -append $output_file {echo $vlog_timestamp};

  # Dump the CVS ID's for all elements in the srcline list
  foreach srcline_elem $srcline {
    echo $srcline_elem;
    set vlog_srcline "$vlog_comment netlist CVS srcid: $srcline_elem";
    redirect -append $output_file {echo $vlog_srcline};
  }; # end foreach
  redirect -append $output_file {echo $vlog_cvsline};

  # By default, don't write the following files
  if {$out_type == "layout"} {
    
    #######################################################################
    # Writes Synopsys Design Constraints (SDC) V1.1 format
    #######################################################################
    set output_file [file join ${outputs_path} ${design_name}${out_string}.sdc];
    echo "#OUTPUTS: Writing sdc synopsys constraints to ${output_file}";
    write_sdc $output_file;
    
  }; # end if
  
  #######################################################################
  # Finished writing outputs
  #######################################################################
  echo "#OUTPUTS: Finished executing P_proj_outputs";
  
}; # end of proc P_proj_outputs

define_proc_attributes P_proj_outputs \
  -info "PROJ_PROCS: Procedure to generate default outputs This includes the db, sdf, \
sdc & verilog gate netlist" \
  -define_args {
    {dbs_path "Path to the dbs file" dbs_path string required}
    {design_name "Name of the design" design_name string required}
    {outputs_path "Path to the output files" outputs_path string required}
    {srcline "Source code revision information" srcline string required}
    {out_string "Output filename string" out_string string}
    {out_type "Output types string" out_type string}
}  

#######################################################################
# Usage: P_proj_ungroup_dw
#
# This procedure is used to ungroup the DesignWare components. This is
# useful to remove unconnected ports (caused layout issues)
#######################################################################
proc P_proj_ungroup_dw {} {
  global G_TOPS;
  global current_design;

  echo "#UNGROUP_DW: ungrouping designware parts"

  # Save current location
  set this_design $current_design;

  # Need to traverse through hierarchy & flatten DW components
  redirect $G_TOPS(NULL) {set hier_designs [filter [find design *] {@is_hierarchical == true}]};
  if {$hier_designs != {}} {
    foreach_in_collection tmp_design $hier_designs {
      redirect $G_TOPS(NULL) {set current_design $tmp_design};
      redirect $G_TOPS(NULL) {set cell_list [filter [find cell *] {@DesignWare == true}]};
      if {$cell_list != {}} {
         foreach_in_collection tmp_cell $cell_list {
           echo "#UNGROUP_DW: ungrouping" [get_object_name $tmp_cell] \
             "in design" [get_object_name $tmp_design];
           ungroup $tmp_cell -flatten -prefix DW_ -simple_names;
	 }; # end foreach_in_collection
      }; # end if
      unset cell_list;
   }; # end foreach_in_collection
  unset tmp_design;
  }; # end if

  # Reset current back to where we started from */
  redirect $G_TOPS(NULL) {set current_design $this_design}

  # Clean up namespace
  unset this_design
  unset hier_designs
}; # end P_proj_ungroup_dw

define_proc_attributes P_proj_ungroup_dw \
  -info "PROJ_PROCS: Used to ungroup the Design Ware components for the design"

#######################################################################
# Usage: P_proj_wire_load 
#
# This procedure is used to set the default wireload for the target. 
# Assumes G_DEFAULT_WIRE_LOAD set from proj_library script
#######################################################################
proc P_proj_wire_load {} {
  global G_DEFAULT_WIRE_LOAD;

  echo "#PROJ: Setting wireload";

  set_wire_load_model -name $G_DEFAULT_WIRE_LOAD;
  set_wire_load_mode top;
  set auto_wire_load_selection false;
}; # end P_proj_wire_load

define_proc_attributes P_proj_wire_load \
  -info "PROJ_PROCS: Sets the default wire load model"

#######################################################################
# $Log$
#######################################################################
