#######################################################################
# File Name    : util_procs.tcl
# Date Created : 12/06/99 
# Author       : Tim L Wilson, Gregg D. Lahti & Rodney Pesavento
# Project      : TOPS Example Code for SNUG
#######################################################################
# $Id$
#######################################################################
# Description: TOPS utility procedures for synthesis
#######################################################################

# Display the TOPS environment release
set G_TOPS_VER 1.5;

echo "#SETUP: Sourcing TOPS procedure file ($G_PROJ_SYN/util_procs.tcl)";
echo "#SETUP: TOPS Environment Version: $G_TOPS_VER\n";

#######################################################################
# Usage: P_elapsed_time mark
# This procedure sets up a global var to allow procs to mark time of scripts
#######################################################################
proc P_elapsed_time {mark} {
  set now [clock seconds];
  clock format [expr $now - $mark -(17*3600)] -format %H:%M:%S;

}; # end elapsed_time;

define_proc_attributes P_elapsed_time \
  -info "TOPS: Set up a global variable to allow procedures to mark time of scripts" \
  -define_args {
    {mark "mark is the time mark" mark float required}
}

#######################################################################
# Usage: P_timestamp
# This procedure prints the timestamp into the log file
#######################################################################
proc P_timestamp {} {
  set c [clock format [clock seconds]];
  set h [sh hostname];
  set u [sh whoami];
  echo "#TIMESTAMP: " $c " " $h " " $u;

}; # end P_timestamp;

define_proc_attributes P_timestamp \
  -info "TOPS: Puts the unix time stamp into the log files"

#######################################################################
# Usage: P_get_srcline filename
# This procedure to get the Id CVS string from the source file.
# NOTE: filename: a fully qualified path and filename or a path and 
#   filename relative to the unit's syn dir.
#######################################################################
proc P_get_srcline {fn sfn} {
  echo "#INFO: (P_get_srcline) : Extracting revision data from file(s) " $fn $sfn;

  # Initialize file list & srcline list variables
  set file_list [concat $fn $sfn];
  set srcline_list "";

  foreach src_file $file_list {
    set srcline [exec sed -n -e "/Id: /p" $src_file];

    # Check if srcline set (doesn't get set when file doesn't have CVS header
    if {$srcline == ""} {
      echo "#WARNING: No extracted revision data from file (please append header template)";
      set srcline "No CVS Id for $src_file (please append header template)"
    }
    regexp {[\$](.+)[\$]} $srcline srcline srcline;
    lappend srcline_list $srcline; 
  }; # end foreach

  return $srcline_list;
}; # end P_get_srcline

define_proc_attributes P_get_srcline \
  -info "TOPS: Procedure to get the Id CVS string from the source file" \
  -define_args {
    {fn "a fully qualified path and filename or a path and filename relative to the unit's syn dir" fn string required}
    {sfn "list of fully qualified subfub files" sfn string required}
}

#######################################################################
# Usage: P_source_if_exists filename caller
# This procedure is used to see if a file exists.  If it doesn't, return a zero.
# caller added to print invoking script's name 
#######################################################################
proc P_source_if_exists {filename caller} {
  if {[which $filename] != ""} {
    set LOCAL_TIME_MARK [clock seconds];
    set full_name [which $filename];
    echo "#$caller: Sourcing $full_name";
    # Source the file in the top-level context
    uplevel source $filename;
#    echo "#$caller: Elapsed Time for Sourcing $full_name: [P_elapsed_time $LOCAL_TIME_MARK]";
    return 1;
  } else {
    # File was not found
    return 0;
  }
}; # end P_source_if_exists

define_proc_attributes P_source_if_exists \
  -info "TOPS: Procedure to source file if it exists" \
  -define_args {
    {filename "File name" filename string required} \
    {caller "Invoking script name" caller string required}
}

#######################################################################
# Usage: P_exec_if_exists filename caller
# Procedure to exec procedure if it exists.  If it doesn't, return
# caller added to print Invoking script's name.
#######################################################################
proc P_exec_if_exists {procname caller} {

  # Test to see if Procedure exists
  if {[info proc ${procname}] == ${procname} } {
    # Procedure exists, execute
    echo "#$caller: Executing ${procname}";
    # Evaluate the proc in the top-level context
    uplevel $procname;
    return 1;
  } else {
    # Procedure does not exist, return
    return 0;
  }
}; # end P_exec_if_exists

define_proc_attributes P_exec_if_exists \
  -info "TOPS: Procedure to execute procedure if it exists" \
  -define_args {
    {procname "Procedure name" procname string required} \
    {caller "Invoking script name" caller string required}
}

#######################################################################
# Usage: P_get_license HDL-Compiler
# This procedure grabs the specified license.  Note that DC gives error status
# if we already have the license.  Hence, must determine if we
# do have the license first before we actually get it.
# Also re-check for the license in 60 second intervals for 
# 1 hour before we exit with an error if we can't get a license.
#######################################################################
proc P_get_license {ln} {

  set fetch_license 1;
  set sleep_val 60;          # sleep in seconds we wait for a license
  set max_timeout 60;        # number of 60-second waits until we die

  # First must determine which licenses we have.  To do this
  # we get the output of list_licenses into a TCL variable.
  #
  # Hack! DC is broke & can't set list_lic output to a variable,
  # so must stuff it to a tmp file & read it back in.  Use process
  # id (pid) in filename as a safer-alternative, touch & rm -rf to
  # first to be extra safe
  exec touch [eval pid].tmp;
  exec rm -rf [eval pid].tmp;
  redirect [eval pid].tmp {list_licenses};
  set licenses [exec cat [eval pid].tmp];
  exec rm -rf [eval pid].tmp;

  # now that we have a TCL variable with the list_license output,
  # strip out the crap that DC uselessly puts in and parse the list
  regsub {Licenses in use:} $licenses {} licenses;
  foreach checkedout_license $licenses {
     if [string match $checkedout_license $ln] {
        set fetch_license 0;
     }
  }
  if {$fetch_license == 1} {
    set sleep_time 0;
    set dc_status [get_license $ln];
    while { ($dc_status == 0) && ($sleep_time < $max_timeout) } {
      if {$sleep_time == 0} {
        set current_time [exec date];
        echo "#INFO: waiting for license $ln @ $current_time";
      }
      sh sleep $sleep_val;
      set sleep_time [expr $sleep_time + 1];
      set dc_status [get_license $ln];
    }
    if {$dc_status == 0} {
      echo "Error: Cannot get license $ln after $sleep_time seconds, dying!";
      return 0;
    } else {
      #echo "#INFO: checked out license $ln";
      return 1;
    }
  } else {
   #echo "#INFO: already have license, continuing along";
   return 1;
  }
}; # end P_get_license

define_proc_attributes P_get_license \
  -info "TOPS: Procedure to get a license" \
  -define_args {
    {ln "license name" ln string required}
}

#######################################################################
# Usage: P_check4lic MOTIVE-PrimeTime
# This procedure is used to check for a particular license. 
# Returns a 1 if the license is checked out.
# Returns a 0 if not checked out. 
# Can be used to determine if PT or DC is running. 
#######################################################################
proc P_check4lic {ln} {
  set license_exists 0;

  # First must determine which licenses we have.  To do this
  # we get the output of list_licenses into a TCL variable.
  #
  # Hack! DC is broke & can't set list_lic output to a variable,
  # so must stuff it to a tmp file & read it back in.  Use process
  # id (pid) in filename as a safer-alternative, touch & rm -rf to
  # first to be extra safe
  exec touch [eval pid].tmp;
  exec rm -rf [eval pid].tmp;
  redirect [eval pid].tmp {list_licenses};
  set licenses [exec cat [eval pid].tmp];
  exec rm -rf [eval pid].tmp;

  # Now that we have a TCL variable with the list_license output,
  # strip out the crap that DC uselessly puts in and parse the list
  regsub {Licenses in use:} $licenses {} licenses;
  foreach checkedout_license $licenses {
     if [string match $checkedout_license $ln] {
        set license_exists 1;
     }
  }
 return $license_exists;
}; # end P_check4lic

define_proc_attributes P_check4lic \
  -info "TOPS: Procedure to check for a particular license" \
  -define_args {
    {ln "license name" ln string required}
}

#######################################################################
# Usage: P_read_vhdl filename.vhd 
# This procedure is used to read the VHDL source code file.
# Library may be optionally specified.
#######################################################################
proc P_read_vhdl {filename {libname default}} {
  P_get_license "VHDL-Compiler";
  if {$libname == "default"} {
    analyze -f vhdl $filename;
  } else {
    analyze -library $libname -f vhdl $filename;
  }
  remove_lic "VHDL-Compiler";
}; # end P_read_vhdl;

define_proc_attributes P_read_vhdl \
  -info "TOPS: Procedure to read a vhdl file" \
  -define_args {
    {filename "VHDL File Name" filename string required}
    {libname "Library Name" libname string}
}

#######################################################################
# Usage: P_read_verilog filename.v
# This procedure is used to read the Verilog source code file.
# Library may be optionally specified.
#######################################################################
proc P_read_verilog {filename {libname default}} {
  P_get_lic "HDL-Compiler";
  if {$libname == "default"} {
    analyze -f verilog $filename;
  } else {
    analyze -library $libname -f verilog $filename;
  }
  remove_lic "HDL-Compiler";
}; # end P_read_verilog;
   
define_proc_attributes P_read_verilog \
  -info "TOPS: Procedure to read a verilog file" \
  -define_args {
    {filename "Verilog File Name" filename string required}
    {libname "Library Name" libname string}
}

#######################################################################
# Usage: P_read_top top_unit
# This procedure is used to read in top unit (language neutral).
# Other file extensions may be added as required
#######################################################################
proc P_read_top {filename {libname default}} {
  set design_name [file tail [file rootname $filename]];
  echo "#INFO: P_read_top: reading $filename";
  switch -exact -- [file extension $filename] {
    .vhd {
      if {$libname == "default"} {
        P_read_vhdl $filename;
        elaborate $design_name;
      } else {
        P_read_vhdl $filename $libname;
        elaborate -library $libname $design_name;
      }
      remove_license VHDL-Compiler;
    }
    .vhdl {
      if {$libname == "default"} {
        P_read_vhdl $filename;
        elaborate $design_name;
      } else {
        P_read_vhdl $filename $libname;
        elaborate -library $libname $design_name;
      }
      remove_license VHDL-Compiler;
    }
    .v {
      if {$libname == "default"} {
        P_read_verilog $filename;
        elaborate $design_name;
      } else {
        P_read_verilog $filename $libname;
        elaborate -library $libname $design_name;
      }
      remove_license HDL-Compiler;
    }
    .db {
      echo "#INFO: Reading db $filename";
      read_db $filename;
      elaborate $design_name;
    }
    default {
      echo "#INFO: Reading db $filename";
      read_db $filename;
      elaborate $design_name;
    }
  }
}; # end P_read_top;

define_proc_attributes P_read_top \
  -info "TOPS: Procedure to read in the target" \
  -define_args {
    {filename "File Name of the target" filename string required}
    {libname "Library Name for the target" libname string}
}

#######################################################################
# Usage: P_analyze_pkg filename.vhd 
# This procedure is used to do analyze only packages for TOPS.
# This is useful for working with VHDL packages.
#######################################################################
proc P_analyze_pkg {fn} {
  global G_PROJ_SYN;
  global synthetic_library;
  global link_library;
  global synopsys_program_name;

  # Load the library setup script
  set myfile [file join ${G_PROJ_SYN} proj_library.tcl];
  P_source_if_exists "$myfile" "SETUP";

  P_get_license "VHDL-Compiler";
  analyze -f vhdl $fn;
  remove_license "VHDL-Compiler";

  # Done, exit code
  exit;
}; # end P_analyze_pkg;

define_proc_attributes P_analyze_pkg \
  -info "TOPS: Procedure to analyze VHDL packages" \
  -define_args {
    {filename "vhdl filename" filename string required}
}

#######################################################################
# Usage: P_read_subfubs $SUBFUBS_list
# This procedure is used to deal with sub-units (subfubs) by
# reading them in rather than compiling them and then reading them
# in.  This is to promote top-down compiles with sub-units.  Used
# in the proj_main.tcl script.
# Project may add extensions as required.
#######################################################################
proc P_read_subfubs {subfubs {libname default}} {
  global G_SRC_PATH;

  # Check if any subfubs to read. If none, exit else process list;
  if {![string match $subfubs ""]} {
    set subfub_list [join $subfubs];     # must join it into list first;
    echo "SUBFUBS: $subfub_list";

    # Now parse out each file;
    foreach {subfub} $subfub_list {
      set dirpath [file dirname $subfub];      # get directory path first;

      # No directory, use $G_SRC_PATH
      if {[string match $dirpath ""] == 1} {
        set subfub [file join $G_SRC_PATH $subfub];
      }

      switch -exact -- [file extension $subfub] {
        .vhd {
          if {$libname == "default"} {
            P_read_vhdl $subfub;
          } else {
            P_read_vhdl $subfub $libname; 
          }
        }
        .vhdl {
          if {$libname == "default"} {
            P_read_vhdl $subfub;
          } else {
            P_read_vhdl $subfub $libname; 
          }
        }
        .v {
          if {$libname == "default"} {
            P_read_verilog $subfub;
          } else {
            P_read_verilog $subfub $libname; 
          }
        }
        .db {
           echo "#INFO: Reading db $subfub";
          read_db $subfub;
        }
        default {
          echo "#INFO: Reading db $subfub";
          read_db $subfub;
        }
      }; # end switch
    }; # end foreach
  }; # end if
}; # end P_read_subfubs;

define_proc_attributes P_read_subfubs \
  -info "TOPS: Procedure to read the subfubs before compiling" \
  -define_args {
    {subfubs "list of the subfubs" subfubs string required}
    {libname "Library Name for the subfubs" libname string}
}

#######################################################################
# Usage: P_compilelib libname libsrcdir
# This procedure is used to compile a library of [libname] into the ./dbs 
# directory from the source pointed to by libsrcdir.  Called from
# Makefile directly (dc_shell-tcl -x P_compilelib [libname libsrc])
#######################################################################
proc P_compilelib {libname libsrcdir} {
	
  set mark_main [clock seconds];  # save start time
  P_timestamp;
  if { $libname == {} } {
    echo {ERROR: variable libname is not set}
    # exit 1
  }
  if { $libsrcdir == {} } {
    echo {ERROR: variable libsrcdir is not set}
    # exit 1
  }
  P_get_license "Design-Compiler"
	
  set myfile [[file join ${libsrcdir} ${libname}.lib];
  echo "#P_compilelib: reading $myfile\n";
  read_lib "$myfile";
	
  set myfile [[file join dbs ${libname}.db];
  echo "#P_compilelib: writing out $myfile";
  write_lib $libname -f db -o "$myfile"
	
  set myfile [file join log ${libname}.hst];
  redirect "$myfile" {history -h};
	
  # done!
  echo "#P_compilelib: Elapsed time of execution was " [P_elapsed_time $mark_main];
  P_timestamp;
}; # end P_compilelib;

define_proc_attributes P_compilelib\
  -info "TOPS: Procedure to compile a library into the ./dbs from the source pointed to by libsrcdir" \
  -define_args {
    {libname "Name of the library to be compiled" libname string required}
    {libsrcdir "Directory that points to the library" libsrcdir string required}
}

#######################################################################
# Usage: P_log_filter {log_path unitname}
# This procedure is used to filter the synthesis log file for the information
# of interest. Errors, warnings & elapsed times.
#######################################################################
proc P_log_filter {log_path design} {   
  set log_file [file join $log_path ${design}.log];
  set filter_log_file [file join $log_path ${design}.log.summary];
  set error_cnt 0; #number of error messages
  set warning_cnt 0; #number of warning messagess   
  set line_cnt 0; #number of line in log file 
  set elapsed_cnt 0; #number of elapsed messages  
  
  set LOG_HANDLE [open $log_file r];
  set FILT_LOG_HANDLE [open $filter_log_file w];
  
  # Search for lines with errors and warnings
  while {[gets $LOG_HANDLE line] >= 0} {
    incr line_cnt 1;
    if {[regexp {^.*Error} $line]} {
      regsub Error $line "" new_line;
      set errors($error_cnt) "line no ${line_cnt}${new_line}";
      incr error_cnt 1;
    } elseif {[regexp {^Warning} $line]} {
      regsub Warning $line "" new_line;
      set warnings($warning_cnt) "line no ${line_cnt}${new_line}";
      incr warning_cnt 1;
    } elseif {[regexp {Elapsed} $line]} {
      set elapsed($elapsed_cnt) $line;
      incr elapsed_cnt 1;
    }
  }
  
  # Print Errors and Lines in a filtered log
  puts $FILT_LOG_HANDLE "There are $error_cnt errors, $warning_cnt warnings";

  # Print the Elapsed messages if any
  if {$elapsed_cnt > 0} {
    puts $FILT_LOG_HANDLE "\nElapsed :";
    for {set i 0} {$i < $elapsed_cnt} {incr i} {
      puts $FILT_LOG_HANDLE $elapsed($i)
    }
  }
 
  # Print the errors if any 
  if {$error_cnt > 0} {                 
    puts $FILT_LOG_HANDLE "\nErrors :";
    for {set i 0} {$i < $error_cnt} {incr i} {
      puts $FILT_LOG_HANDLE $errors($i);
    }
  }

  # Print the warnings if any
  if {$warning_cnt > 0} {
    puts $FILT_LOG_HANDLE "\nWarnings :";
    for {set i 0} {$i < $warning_cnt} {incr i} {
      puts $FILT_LOG_HANDLE $warnings($i);
    }
  }

  close $FILT_LOG_HANDLE;
  close $LOG_HANDLE;
}; # end proc P_log_filter

define_proc_attributes P_log_filter \
  -info "TOPS: Procedure to filter the logfile and print only the errors, warnings and elapsed messages" \
  -define_args {
    {log_path "the directory in which the log_file is to be created" log_path string required}
    {design "Name of the unit" design string required}
}

#######################################################################
# $Log$
#######################################################################



